"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var fs = require("fs");
exports.Kinds = ts.SyntaxKind;
var TSHost = /** @class */ (function () {
    function TSHost(rootFileNames) {
        var _this = this;
        this.rootFileNames = rootFileNames;
        this.files = {};
        this.fileExists = ts.sys.fileExists;
        this.readFile = ts.sys.readFile;
        this.readDirectory = ts.sys.readDirectory;
        rootFileNames.forEach(function (fileName) {
            _this.files[fileName] = { version: 0 };
        });
    }
    TSHost.prototype.getScriptFileNames = function () {
        return this.rootFileNames;
    };
    TSHost.prototype.getScriptVersion = function (fileName) {
        return this.files[fileName] && this.files[fileName].version.toString();
    };
    TSHost.prototype.getScriptSnapshot = function (fileName) {
        if (!fs.existsSync(fileName)) {
            return undefined;
        }
        return ts.ScriptSnapshot.fromString(fs.readFileSync(fileName).toString());
    };
    TSHost.prototype.getCurrentDirectory = function () {
        return process.cwd();
    };
    TSHost.prototype.getCompilationSettings = function () {
        return {};
    };
    TSHost.prototype.getDefaultLibFileName = function (options) {
        return ts.getDefaultLibFilePath(options);
    };
    return TSHost;
}());
exports.TSHost = TSHost;
;
var TSProject = /** @class */ (function () {
    function TSProject(files) {
        this.files = files;
        this.host = new TSHost(this.files);
        this.language = ts.createLanguageService(this.host, ts.createDocumentRegistry());
        this.program = this.language.getProgram();
    }
    TSProject.prototype.getFile = function (name) {
        var source = this.program.getSourceFile(name);
        if (!source)
            return null;
        return new TSFile(this, source);
    };
    return TSProject;
}());
exports.TSProject = TSProject;
var TSFile = /** @class */ (function () {
    function TSFile(project, source) {
        this.project = project;
        this.source = source;
    }
    TSFile.prototype.getKindTree = function (node, depth) {
        var _this = this;
        if (depth === void 0) { depth = 1; }
        var tree = {
            name: this.getSyntaxKindName(node.kind),
            children: []
        };
        if (depth === 0) {
            tree.children = null;
            return tree;
        }
        ts.forEachChild(node, function (child) {
            tree.children.push(_this.getKindTree(child, depth - 1));
        });
        return tree;
    };
    TSFile.prototype.inspectKind = function (node, depth) {
        if (depth === void 0) { depth = 3; }
        var tree = this.getKindTree(node, depth);
        this.printKindTree(tree);
    };
    TSFile.prototype.getSyntaxKindName = function (kind) {
        for (var name_1 in ts.SyntaxKind) {
            if (ts.SyntaxKind[name_1] === kind) {
                return name_1;
            }
        }
        return "Unknown";
    };
    TSFile.prototype.printKindTree = function (tree, indent) {
        if (indent === void 0) { indent = ""; }
        console.error(indent + tree.name);
        for (var _i = 0, _a = tree.children || []; _i < _a.length; _i++) {
            var child = _a[_i];
            this.printKindTree(child, indent + "  ");
        }
    };
    TSFile.prototype.getClasses = function (node) {
        var _this = this;
        if (node === void 0) { node = this.source; }
        var specs = [];
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.ClassDeclaration) {
                specs.push(_this.parseClass(child));
            }
            else {
                specs.push.apply(specs, _this.getClasses(child));
            }
        });
        return specs;
    };
    TSFile.prototype.getClass = function (name, node) {
        if (node === void 0) { node = this.source; }
        var classes = this.getClasses(node);
        for (var _i = 0, classes_1 = classes; _i < classes_1.length; _i++) {
            var cla = classes_1[_i];
            if (cla.name == name)
                return cla;
        }
        return null;
    };
    TSFile.prototype.getNamespaces = function (node) {
        var _this = this;
        if (node === void 0) { node = this.source; }
        var specs = [];
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.ModuleDeclaration) {
                specs.push(_this.parseNamespace(child));
            }
            else {
                //this.inspectKind(node, 5)
                specs.push.apply(specs, _this.getNamespaces(child));
            }
        });
        return specs;
    };
    TSFile.prototype.getNamespace = function (name, node) {
        if (node === void 0) { node = this.source; }
        var namespacees = this.getNamespaces(node);
        for (var _i = 0, namespacees_1 = namespacees; _i < namespacees_1.length; _i++) {
            var ns = namespacees_1[_i];
            if (ns.name == name)
                return ns;
        }
        return null;
    };
    TSFile.prototype.parseClass = function (node) {
        var _this = this;
        var spec = {
            name: null,
            constructor: null,
            methods: {},
            properties: {},
            generics: [],
            base: null,
            node: node,
            children: []
        };
        ts.forEachChild(node, function (child) {
            spec.children.push(node);
            if (child.kind == exports.Kinds.ExportKeyword) {
                spec.export = true;
            }
            else if (child.kind == exports.Kinds.AbstractKeyword) {
                spec.abstract = true;
            }
            else if (child.kind == exports.Kinds.Identifier) {
                var id = child;
                spec.name = id.text;
            }
            else if (child.kind == exports.Kinds.Constructor) {
                var cons = child;
                var info = _this.parseMethod(child);
                spec.constructor = info;
            }
            else if (child.kind == exports.Kinds.MethodDeclaration) {
                var method = child;
                var info = _this.parseMethod(child);
                spec.methods[info.name] = info;
            }
            else if (child.kind == exports.Kinds.PropertyDeclaration) {
                var prop = child;
                var info = _this.parseProperty(child);
                spec.properties[info.name] = info;
            }
            else if (child.kind == exports.Kinds.HeritageClause) {
                var base = child;
                spec.base = _this.parseHeritage(base);
            }
            else if (child.kind == exports.Kinds.TypeParameter) {
                spec.generics.push(_this.parseTypeParameter(child));
            }
            else {
                console.trace();
                console.error("Unknown kind", _this.getSyntaxKindName(child.kind), _this.getSyntaxKindName(node.kind));
                console.error(child.getText());
                console.error(node.getText());
                _this.inspectKind(child, 8);
            }
        });
        return spec;
    };
    TSFile.prototype.parseTypeParameter = function (node) {
        var spec = {
            name: null,
            default: null
        };
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.Identifier) {
                var id = child;
                spec.name = id.text;
            }
        });
        return spec;
    };
    TSFile.prototype.parseFunction = function (node) {
        var _this = this;
        var spec = {
            parameters: [],
            node: node,
            block: null
        };
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.Identifier) {
                var id = child;
                spec.name = id.text;
            }
            else if (child.kind == exports.Kinds.Parameter) {
                var param = child;
                var info = _this.parseParam(param);
                spec.parameters.push(info);
            }
            else {
                spec.block = _this.parseBlock(child);
            }
        });
        return spec;
    };
    TSFile.prototype.parseMethod = function (node) {
        var _this = this;
        var spec = {
            name: null,
            parameters: [],
            permission: "public",
            node: node,
            block: null
        };
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.PublicKeyword) {
                spec.permission = "public";
            }
            else if (child.kind == exports.Kinds.PrivateKeyword) {
                spec.permission = "private";
            }
            else if (child.kind == exports.Kinds.ProtectedKeyword) {
                spec.permission = "protected";
            }
            else if (child.kind == exports.Kinds.Identifier) {
                var id = child;
                spec.name = id.text;
            }
            else if (child.kind == exports.Kinds.Parameter) {
                var param = child;
                var info = _this.parseParam(param);
                spec.parameters.push(info);
            }
            else {
                spec.block = _this.parseBlock(child);
            }
        });
        return spec;
    };
    TSFile.prototype.parseProperty = function (node) {
        var _this = this;
        var spec = {
            name: null,
            permission: "public",
            node: node,
        };
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.AbstractKeyword) {
                spec.abstract = true;
            }
            else if (child.kind == exports.Kinds.Identifier) {
                var id = child;
                spec.name = id.text;
            }
            else if (child.kind == exports.Kinds.ReadonlyKeyword) {
                spec.readonly = true;
            }
            else if (child.kind == exports.Kinds.PublicKeyword) {
                spec.permission = "public";
            }
            else if (child.kind == exports.Kinds.PrivateKeyword) {
                spec.permission = "private";
            }
            else if (child.kind == exports.Kinds.ProtectedKeyword) {
                spec.permission = "protected";
            }
            else if (child.kind == exports.Kinds.TypeLiteral) {
                spec.type = _this.parseTypeLiteral(child);
            }
            else if (child.kind == exports.Kinds.TypeReference) {
                spec.type = _this.parseTypeReference(child);
            }
            else if (child.kind == exports.Kinds.FunctionExpression) {
                spec.type = _this.parseTypeReference(child);
            }
            else if (child.kind == exports.Kinds.ArrayType) {
                var arr = child;
                spec.type = _this.parseTypeReference(arr);
            }
            else {
                spec.value = _this.parseValue(child);
            }
        });
        return spec;
    };
    TSFile.prototype.parseParam = function (node) {
        var _this = this;
        var spec = {
            name: null,
            type: null,
            node: node,
        };
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.Identifier) {
                var id = child;
                spec.name = id.text;
            }
            else if (child.kind == exports.Kinds.TypeReference) {
                var ref = child;
                spec.type = _this.parseTypeReference(ref);
            }
            else if (child.kind == exports.Kinds.TypeLiteral) {
                var ref = child;
                spec.type = _this.parseTypeLiteral(ref);
            }
        });
        return spec;
    };
    TSFile.prototype.parseTypeLiteral = function (node) {
        var spec = {
            literal: null,
            base: null,
            reference: null,
            genericLiteral: null,
            node: node,
        };
        spec.literal = node.getText(this.source);
        return spec;
    };
    TSFile.prototype.parseTypeReference = function (node) {
        var spec = {
            literal: null,
            base: null,
            reference: null,
            genericLiteral: null,
            node: node,
        };
        spec.literal = node.getText(this.source);
        return spec;
    };
    TSFile.prototype.parseNamespace = function (node) {
        var _this = this;
        var spec = {
            name: null,
            node: node,
            block: null,
        };
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.ExportKeyword) {
                spec.export = true;
            }
            else if (child.kind == exports.Kinds.Identifier) {
                var id = child;
                spec.name = id.text;
            }
            else if (child.kind == exports.Kinds.ModuleBlock) {
                spec.block = _this.parseBlock(child);
            }
            else {
                console.error("Unknown name space kind", _this.getSyntaxKindName(child.kind));
            }
        });
        return spec;
    };
    TSFile.prototype.parseBlock = function (node) {
        var _this = this;
        if (node === void 0) { node = this.source; }
        var spec = {
            node: node,
            variables: {},
            classes: {},
            namespaces: {},
            children: [],
        };
        ts.forEachChild(node, function (child) {
            spec.children.push(child);
            if (child.kind == exports.Kinds.VariableStatement) {
                var vars = _this.parseVariableStatement(child);
                for (var _i = 0, vars_1 = vars; _i < vars_1.length; _i++) {
                    var va = vars_1[_i];
                    spec.variables[va.name] = va;
                }
            }
            else if (child.kind == exports.Kinds.ModuleDeclaration) {
                var mod = child;
                var modSpec = _this.parseNamespace(mod);
                spec.namespaces[modSpec.name] = modSpec;
            }
            else if (child.kind == exports.Kinds.ClassDeclaration) {
                var claSpec = _this.parseClass(child);
                spec.classes[claSpec.name] = claSpec;
            }
        });
        return spec;
    };
    TSFile.prototype.parseVariableStatement = function (node) {
        var _this = this;
        var specs = [];
        var exp = false;
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.ExportKeyword) {
                exp = true;
            }
            else if (child.kind == exports.Kinds.VariableDeclarationList) {
                var list = child;
                for (var _i = 0, _a = list.declarations; _i < _a.length; _i++) {
                    var dec = _a[_i];
                    var spec = _this.parseVariableDeclaration(dec);
                    spec.export = exp;
                    specs.push(spec);
                }
            }
        });
        return specs;
    };
    TSFile.prototype.parseVariableDeclaration = function (node) {
        var _this = this;
        var id = node.exclamationToken;
        var spec = {
            name: null,
            value: null
        };
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.Identifier) {
                var id_1 = child;
                spec.name = id_1.text;
            }
            else {
                spec.value = _this.parseValue(child);
            }
        });
        return spec;
    };
    TSFile.prototype.parseStatement = function (node) {
        var spec = {
            type: null,
        };
        if (node.kind == exports.Kinds.VariableStatement) {
            var vs = node;
            ts.forEachChild(vs, function (child) {
                if (child.kind == exports.Kinds.ExportKeyword) {
                    spec.export = true;
                }
                else if (child.kind == exports.Kinds.VariableDeclarationList) {
                    var list = child;
                    list.declarations;
                }
            });
        }
        return spec;
    };
    TSFile.prototype.explain = function (depth) {
        if (depth === void 0) { depth = 10; }
        this.inspectKind(this.source, depth);
    };
    TSFile.prototype.parseHeritage = function (node) {
        var _this = this;
        var spec = {
            name: null,
            class: null,
            node: node,
            generics: []
        };
        var id = null;
        // We may find many identifier, we use the last one
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.ExpressionWithTypeArguments) {
                var exp = child;
                ts.forEachChild(child, function (child) {
                    if (child.kind == exports.Kinds.PropertyAccessExpression) {
                        var acc = _this.resolvePropertyAccess(child);
                        var last = acc.nodes[acc.nodes.length - 1];
                        if (last) {
                            id = last;
                            spec.name = id.text;
                        }
                        else {
                            console.error("Unknown access expression");
                            _this.inspectKind(child, 3);
                        }
                    }
                });
            }
            else if (child.kind == exports.Kinds.PropertyAccessExpression) {
                var acc = _this.resolvePropertyAccess(child);
                var last = acc.nodes[acc.nodes.length - 1];
                if (last) {
                    id = last;
                    spec.name = id.text;
                }
                else {
                    console.error("Unknown access expression");
                    _this.inspectKind(child, 3);
                }
            }
        });
        if (!id) {
            //this.inspectKind(node, 8)
            //console.log("Missing hist id", node.getText())
        }
        if (id) {
            spec.name;
            var def = this.project.language.getDefinitionAtPosition(this.source.fileName, id.getStart());
            var checker = this.project.program.getTypeChecker();
            var symbol = checker.getSymbolAtLocation(id);
            var type = checker.getTypeOfSymbolAtLocation(symbol, id);
            if (type && type.symbol) {
                var cla = type.symbol.declarations[0];
                spec.class = this.parseClass(cla);
            }
            else {
                console.log("Missing symbol", id.text);
            }
        }
        return spec;
    };
    TSFile.prototype.resolve = function (symbol) {
        var node = symbol.declarations[0];
        var checker = this.project.program.getTypeChecker();
        return checker.getSymbolAtLocation(node);
    };
    TSFile.prototype.shouldSkipAlias = function (node, declaration) {
        if (node.kind !== exports.Kinds.Identifier) {
            return false;
        }
        if (node.parent === declaration) {
            return true;
        }
        switch (declaration.kind) {
            case exports.Kinds.ImportClause:
            case exports.Kinds.ImportEqualsDeclaration:
                return true;
            case exports.Kinds.ImportSpecifier:
                return declaration.parent.kind === exports.Kinds.NamedImports;
            default:
                return false;
        }
    };
    TSFile.prototype.resolvePropertyAccess = function (node) {
        var _this = this;
        var spec = {
            accesses: [],
            nodes: []
        };
        ts.forEachChild(node, function (child) {
            if (child.kind == exports.Kinds.PropertyAccessExpression) {
                spec = _this.resolvePropertyAccess(child);
            }
            else if (child.kind == exports.Kinds.Identifier) {
                var id = child;
                spec.accesses.push(id.text);
                spec.nodes.push(id);
            }
            else if (child.kind == exports.Kinds.ThisKeyword) {
                spec.accesses.push("this");
                spec.nodes.push(child);
            }
            else if (child.kind == exports.Kinds.SuperKeyword) {
                spec.accesses.push("super");
                spec.nodes.push(child);
            }
            else if (child.kind == exports.Kinds.CallExpression) {
                spec.accesses.push(child.getFullText());
                spec.nodes.push(child);
                spec.anonymous = true;
            }
            else {
                console.error("Unknown access element");
                _this.inspectKind(node);
            }
        });
        return spec;
    };
    TSFile.prototype.parseValue = function (node) {
        var value = {
            type: "literal",
            node: node
        };
        if (node.kind == exports.Kinds.StringLiteral) {
            var str = node;
            value.literal = str.getText();
            value.literalType = "string";
            value.literalValue = JSON.parse(value.literal);
        }
        else if (node.kind == exports.Kinds.NumericLiteral) {
            var num = node;
            value.literal = num.text;
            value.literalType = "number";
            value.literalValue = JSON.parse(num.text);
        }
        else if (node.kind == exports.Kinds.TrueKeyword) {
            var yes = node;
            value.literal = "true";
            value.literalType = "boolean";
            value.literalValue = true;
        }
        else if (node.kind == exports.Kinds.FalseKeyword) {
            var no = node;
            value.literal = "false";
            value.literalType = "boolean";
            value.literalValue = false;
        }
        else if (node.kind == exports.Kinds.ObjectLiteralExpression) {
        }
        else if (node.kind == exports.Kinds.PropertyAccessExpression) {
            value.type = "reference";
            value.access = this.resolvePropertyAccess(node);
        }
        else if (node.kind == exports.Kinds.CallExpression) {
            value.type = "invoke";
            value.invoke = this.parseInvoke(node);
        }
        return value;
    };
    TSFile.prototype.parseInvoke = function (node) {
        var _this = this;
        var spec = {
            node: node,
            invoke: null,
            params: []
        };
        var stack = [];
        ts.forEachChild(node, function (child) {
            stack.push(_this.parseValue(child));
        });
        spec.invoke = stack.shift();
        spec.params = stack;
        return spec;
    };
    TSFile.prototype.querySelect = function (node, route) {
        if (route.length == 0)
            return node;
        if (!node)
            return null;
        if (typeof route == "string") {
            route = [route];
        }
        var first = route.shift();
        var children = node.getChildren();
        for (var _i = 0, children_1 = children; _i < children_1.length; _i++) {
            var child = children_1[_i];
            if (this.getSyntaxKindName(child.kind) == first) {
                return this.querySelect(child, route);
            }
        }
        return null;
    };
    return TSFile;
}());
exports.TSFile = TSFile;
//# sourceMappingURL=tsParser.js.map